# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from abc import ABC
from typing import Any, Dict, List, Sequence

import numpy as np
from pandas import DataFrame

from . import backend as B
from .config import get_accelerator
from .dataclass import ResultCollection
from .embedder import Embedder
from .enums import AcceleratorType, IndexerType
from .indexer import Indexer
from .types import BatchEmbeddings


class UniSim(ABC):
    """Abstract base class for UniSim.

    UniSim is designed to find near-duplicates and similar matches for content
    (such as text, images, multi-modal inputs) using embedding models. The
    UniSim packages enables finding the similarity between two items, indexing
    and querying for similar items, and finding near-duplicates in a dataset.
    """

    def __init__(
        self,
        store_data: bool,
        index_type: str | IndexerType,
        return_embeddings: bool,
        batch_size: int,
        use_accelerator: bool,
        model_id: str,
        embedder: Embedder,
        index_params: Dict[str, Any] | None = None,
        verbose: int = 0,
    ):
        """Create UniSim to index, search, and find near-duplicate matches
        using an embedding model.

        Args:
            store_data: Whether to store a copy of the indexed data and return
                data when returning search results.

            index_type: Indexer type, either "exact" for exact search or
                "approx" for Approximate Nearest Neighbor (ANN) search.

            return_embeddings: Whether to return embeddings corresponding to each
                search result when calling UniSim.search().

            batch_size: Batch size for inference.

            use_accelerator: Whether to use an accelerator (GPU), if available.

            model_id: String id of the model.

            embedder: Embedder object which converts inputs into embeddings using
                a model.

            params: Additional parameters to be passed into the Indexer,
                only used when index_type is approx, for ANN search. Supported
                dict keys include {metric, dtype, connectivity, expansion_add,
                expansion_search}, please see the USearch Index documentation
                for more detail on what each parameter does.

            verbose: Verbosity level, set to 1 for verbose.
        """
        self.store_data = store_data
        self.index_type = index_type if isinstance(index_type, IndexerType) else IndexerType[index_type]
        self.return_embeddings = return_embeddings
        self.batch_size = batch_size
        self.use_accelerator = use_accelerator
        self.model_id = model_id
        self.embedder = embedder
        self.index_params = index_params if index_params else {}
        self.verbose = verbose

        if self.store_data:
            print("INFO: UniSim is storing a copy of the indexed data")
            print("INFO: If you are using large data corpus, consider disabling this behavior using store_data=False")
        else:
            print("INFO: UniSim is not storing a copy of the indexed data to save memory")
            print("INFO: If you want to store a copy of the data, use store_data=True")

        if use_accelerator and get_accelerator() == AcceleratorType.cpu:
            print("INFO: Accelerator is not available, using CPU")
            self.use_accelerator = False

        # internal state
        self.index_size = 0
        self.is_initialized: bool = False
        self.indexed_data: List[Any] = []

    def _lazy_init(self):
        """Lazily init models and indexers."""
        # check we don't initialize indexer twice
        if self.is_initialized:
            return

        # initialize indexer
        self.embedding_size = self.embedder.embedding_size
        self.indexer = Indexer(
            embedding_size=self.embedding_size,
            index_type=self.index_type,
            params=self.index_params,
        )
        self.is_initialized = True

    def embed(
        self,
        inputs: Sequence[Any],
    ) -> BatchEmbeddings:
        """Compute embeddings for inputs.

        Args:
            inputs: Sequence of inputs to embed.

        Returns:
            Embeddings for each input in `inputs`.
        """
        self._lazy_init()
        embeddings = []
        for b_offset in range(0, len(inputs), self.batch_size):
            batch = inputs[b_offset : b_offset + self.batch_size]
            embs = self.embedder.embed(inputs=batch)
            embeddings.append(embs)

        embeddings = np.concatenate(embeddings, axis=0)
        return embeddings

    def similarity(self, input1: Any, input2: Any) -> float:
        """Return the similarity between two inputs as a float.

        Args:
            input1: First input.

            input2: Second input.

        Returns:
            Similarity between the two inputs between 0 and 1.
        """
        # compute embeddings
        batch = [input1, input2]
        embs = self.embed(batch)

        # compute similarity
        similarity = B.cosine_similarity(embs, embs)
        similarity = np.asanyarray(similarity[0][1])

        # clip sometimes for floating point error
        similarity = np.clip(similarity, 0, 1)

        return float(similarity)

    def add(self, inputs: Sequence[Any]) -> List[int]:
        """Add inputs to the index.

        Args:
            inputs: Inputs to embed and add to the index.

        Returns:
             idxs: Indices corresponding to the inputs added to the index,
             where idxs[0] is the idx for inputs[0] in the index.
        """
        inputs_idxs = []
        for b_offset in range(0, len(inputs), self.batch_size):
            batch = inputs[b_offset : b_offset + self.batch_size]
            embs = self.embed(batch)

            # compute the new idxs
            idxs = [i + self.index_size for i in range(len(embs))]
            self.index_size += len(idxs)

            # indexing embeddings
            self.indexer.add(embs, idxs)

            # store inputs if requested
            if self.store_data:
                self.indexed_data.extend(batch)

            # store the idxs corresponding to inputs
            inputs_idxs.extend(idxs)

        return inputs_idxs

    def search(
        self, queries: Sequence[Any], similarity_threshold: float, k: int = 1, drop_closest_match: bool = False
    ) -> ResultCollection:
        """Search for and return the k closest matches for a set of queries,
        and mark the ones that are closer than `similarity_threshold` as
        near-duplicate matches.

        Args:
            queries: Query inputs to search for.

            similarity_threshold: Similarity threshold for near-duplicate
                match, where a query and a search result are considered
                near-duplicate matches if their similarity is higher than
                `similarity_threshold`.

            k: Number of nearest neighbors to lookup for each query input.

            drop_closest_match: If True, remove the closest match before returning
                results. This is used when search queries == indexed set, since
                each query's closest match will be itself if it was already added
                to the index.

        Returns
            result_collection: ResultCollection containing the search results.
        """
        results = ResultCollection()
        for b_offset in range(0, len(queries), self.batch_size):
            batch = queries[b_offset : b_offset + self.batch_size]
            embs = self.embed(batch)

            r = self.indexer.search(
                queries=batch,
                query_embeddings=embs,
                similarity_threshold=similarity_threshold,
                k=k,
                drop_closest_match=drop_closest_match,
                return_data=self.store_data,
                return_embeddings=self.return_embeddings,
                data=self.indexed_data,
            )
            results.merge_result_collection(r)

        return results

    def match(
        self, queries: Sequence[Any], targets: Sequence[Any] | None = None, similarity_threshold: float = 0.9
    ) -> DataFrame:
        """Find the closest matches for queries in a list of targets and
        return the result as a pandas DataFrame.

        Args:
            queries: Input queries to search for.

            targets: Targets to search in, e.g. for each query, find the nearest
                match in `targets`. If None, then `queries` is used as the
                targets as well and matches are computed within a single list.

            similarity_threshold: Similarity threshold for near-duplicate
                match, where a query and a search result are considered
                near-duplicate matches if their similarity is higher than
                `similarity_threshold`.

        Returns:
            Returns a pandas DataFrame with ["query", "target", "similarity", "is_match"]
            columns, representing each query, nearest match in `targets`, their similarity
            value, and whether or not they are a match (if their similarity >=
            `similarity_threshold`).
        """
        # reset index
        self.reset_index()

        # defaults if we have targets
        drop_closest_match = False
        k = 1

        # if we are matching within the same list, drop the closest match
        # since it will be the query itself
        if not targets:
            targets = queries
            drop_closest_match = True
            k = 2

        # add all targets
        self.add(targets)

        # search all queries, so it doesn't depend on similarity threshold
        results = self.search(queries, similarity_threshold=0.0, k=k, drop_closest_match=drop_closest_match).results

        # create pandas df
        data = []
        for i in range(len(results)):
            query = queries[i]
            match = results[i].matches[0]
            similarity = match.similarity
            is_match = similarity and similarity >= similarity_threshold
            matched = targets[match.idx]
            data.append([query, matched, similarity, is_match])

        df = DataFrame(data, columns=["query", "target", "similarity", "is_match"])
        return df

    def reset_index(self):
        """Reset the index by removing all previously-indexed data."""
        self._lazy_init()
        self.index_size = 0
        self.indexer.reset()
        self.indexed_data = []

    def info(self):
        """Display UniSim info."""
        self._lazy_init()

        print("[Embedder]")
        print(f"|-batch_size: {self.batch_size}")
        print(f"|-model_id: {self.model_id}")
        print(f"|-embedding_size: {self.embedding_size}")

        print("[Indexer]")
        print(f"|-index_type: {self.index_type.name}")
        print(f"|-use_accelerator: {self.use_accelerator}")
        print(f"|-store index data: {self.store_data}")
        print(f"|-return embeddings: {self.return_embeddings}")



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, List, Sequence, Union

import numpy as np

from .. import backend as B
from ..types import BatchEmbeddings


class Embedder(ABC):
    """Abstract base class for embedding inputs into vectors."""

    def __init__(self, batch_size: int, model_id: str, verbose: int = 0) -> None:
        """Initialize an Embedder based on a model.

        Args:
            batch_size: Batch size for inference.

            model_id: ID of model.

            verbose: Verbosity mode.
        """
        self.batch_size = batch_size
        self.model_id = model_id
        self.verbose = verbose

        # TODO (marinazh): host on cloud and download when needed instead
        fpath = Path(__file__)
        model_path = fpath.parent / "models" / self.model_id

        if self.verbose:
            print("[Loading model]")
            print(f"|-model_id: {model_id}")

        self.model = B.load_model(model_path, verbose=verbose)

    @property
    @abstractmethod
    def embedding_size(self):
        """Returns the embedding size of the model."""
        return NotImplementedError

    @abstractmethod
    def embed(self, inputs: Sequence[Any]) -> BatchEmbeddings:
        """Compute embeddings for a batch of inputs.

        Args:
            inputs: Sequence of inputs to embed.

        Returns:
            Embeddings for each input in `inputs`.
        """
        raise NotImplementedError

    def predict(self, data: Union[Sequence[Any], np.ndarray]) -> BatchEmbeddings:
        "Run inference using the loaded model with the right framework"
        embeddings: List[np.ndarray] = []
        for idx in range(0, len(data), self.batch_size):
            batch = data[idx : idx + self.batch_size]
            batch_embs = B.predict(self.model, batch=batch)
            embeddings.extend(batch_embs)
        embeddings_array: BatchEmbeddings = np.asanyarray(embeddings)
        return embeddings_array



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from typing import Any, Dict, List, Sequence

from pandas import DataFrame
from tabulate import tabulate

from .dataclass import Result, ResultCollection
from .embedder import TextEmbedder
from .enums import IndexerType
from .types import BatchEmbeddings
from .unisim import UniSim


class TextSim(UniSim):
    """TextSim class for computing textual similarity.

    TextSim leverages the RETSim embedding model to compute the syntactic
    similarity between texts, retrieve similar documents, and deduplicate
    datasets. The model is a highly-optimized, small model that runs
    efficiently on both GPU and CPU, which makes TextSim significantly faster
    than traditional edit distance-based techniques and other neural-based
    embeddings.

    For more details on the RETSim model, please see TODO (marinazh) [paper link].
    """

    def __init__(
        self,
        store_data: bool = True,
        index_type: str | IndexerType = "exact",
        return_embeddings: bool = True,
        batch_size: int = 128,
        use_accelerator: bool = True,
        model_id: str = "text/retsim/v1",
        index_params: Dict[str, Any] | None = None,
        verbose: int = 0,
    ):
        """Create TextSim to index, search, and compute similarity between
        texts using RETSim embedding model.

        Args:
            store_data: Whether to store a copy of the indexed data and return
                data when returning search results. For large datasets
                (e.g. millions of items), we recommend setting store_data=False
                to reduce memory usage.

            index_type: Indexer type, either "exact" for exact search or
                "approx" for Approximate Nearest Neighbor (ANN) search. Using
                exact search guarantees you will return the most similar texts
                to your search queries every single time. We recommend using
                "approx" for large datasets for computational efficiency and it
                is accurate >99% of the time.

            return_embeddings: Whether to return the embeddings corresponding
                to each search result when calling TextSim.search().

            batch_size: Batch size for model inference.

            use_accelerator: Whether to use an accelerator (GPU), if available.

            model_id: String id of the model. Defaults to "text/retsim/v1" for
                the RETSim model.

            index_params: Additional parameters to be passed into the Indexer,
                only used when index_type is "approx" (ANN search). Supported
                dict keys include {dtype, connectivity, expansion_add,
                expansion_search}, please see the USearch Index documentation
                for more details (https://github.com/unum-cloud/usearch).

            verbose: Verbosity level, set to 1 for verbose.
        """
        embedder = TextEmbedder(
            batch_size=batch_size,
            model_id=model_id,
            verbose=verbose,
        )
        super().__init__(
            store_data=store_data,
            index_type=index_type,
            return_embeddings=return_embeddings,
            batch_size=batch_size,
            use_accelerator=use_accelerator,
            model_id=model_id,
            embedder=embedder,
            index_params=index_params,
            verbose=verbose,
        )

    def similarity(self, input1: str, input2: str) -> float:
        """Return the similarity between two texts as a float.

        Args:
            input1: First input.

            input2: Second input.

        Returns:
            Similarity between the two texts, a value between 0 and 1.
        """
        return super().similarity(input1=input1, input2=input2)

    def match(
        self, queries: Sequence[str], targets: Sequence[Any] | None = None, similarity_threshold: float = 0.9
    ) -> DataFrame:
        """Find the closest matches for queries in a list of targets and
        return the result as a pandas DataFrame.

        Args:
            queries: Input query texts to search for.

            targets: Targets to search in, e.g. for each query, find the nearest
                match in `targets`. If None, then `queries` is used as the
                targets as well and matches are computed within a single list.

            similarity_threshold: Similarity threshold for near-duplicate
                match, where a query and a search result are considered
                near-duplicate matches if their similarity is higher than
                `similarity_threshold`. 0.9 is typically a good threshold
                for detecting near-duplicate matches.

        Returns:
            Returns a pandas DataFrame with ["Query", "Match", "Similarity"]
            columns, representing each query, nearest match in `targets`, and
            their similarity value.
        """
        return super().match(queries=queries, targets=targets, similarity_threshold=similarity_threshold)

    def embed(self, inputs: Sequence[str]) -> BatchEmbeddings:
        """Compute embeddings for input texts.

        Args:
            inputs: Sequence of input texts to embed.

        Returns:
            Embeddings for each text in `inputs`. The RETSim model embeds
            each input text into a 256-dimensional float32 array.
        """
        return super().embed(inputs=inputs)

    def add(self, inputs: Sequence[str]) -> List[int]:
        """Embed and add inputs to the index.

        Args:
            inputs: Inputs to embed and add to the index.

        Returns:
             idxs: Indices corresponding to the inputs added to the index,
             where idxs[0] is the idx for inputs[0] in the index.
        """
        return super().add(inputs=inputs)

    def search(
        self, queries: Sequence[str], similarity_threshold: float = 0.9, k: int = 1, drop_closest_match: bool = False
    ) -> ResultCollection:
        """Search for and return the k closest matches for a set of queries,
        and mark the ones that are closer than `similarity_threshold` as
        near-duplicate matches.

        Args:
            queries: Query input texts to search for.

            similarity_threshold: Similarity threshold for near-duplicate
                match, where a query and a search result are considered
                near-duplicate matches if their similarity is higher than
                `similarity_threshold`. 0.9 is typically a good threshold
                for detecting near-duplicate matches.

            k: Number of nearest neighbors to lookup for each query input.

            drop_closest_match: If True, remove the closest match before returning
                results. This is used when search queries == indexed set, since
                each query's closest match will be itself if it was already added
                to the index.

        Returns
            result_collection: ResultCollection containing the search results.
            The total number of near-duplicate matches can be accessed through
            result_collection.total_matches and the list of results for each
            query can be accessed through result_collection.results.
        """
        return super().search(
            queries=queries, similarity_threshold=similarity_threshold, k=k, drop_closest_match=drop_closest_match
        )

    def visualize(self, result: Result, max_display_chars: int = 64):
        """Visualize a search result, displaying the query, closest
        search results and matches, and their similarity values.

        Args:
            result: Result to visualize.

            max_display_chars: Max number of characters to display for text.
        """
        rows = []
        for m in result.matches:
            # check content
            sim = round(m.similarity, 2) if m.similarity else ""
            content = m.data[:max_display_chars] if m.data else ""

            # build row
            row = [m.idx, m.is_match, sim, content]
            rows.append(row)
        if result.query_data:
            print(f'Query {result.query_idx}: "{result.query_data}"')
        else:
            print(f"Query {result.query_idx}")

        print("Most similar matches:\n")

        headers = ["idx", "is_match", "similarity"]
        if self.store_data:
            headers.append("text")

        print(tabulate(rows, headers=headers))



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from .enums import AcceleratorType, BackendType

# internal state
_INTX: str = "int64"
_FLOATX: str = "float32"
_BACKEND: BackendType = BackendType.unknown
_ACCELERATOR: AcceleratorType = AcceleratorType.unknown


def floatx():
    """Returns the default float type, as a string.

    E.g. `'float16'`, `'float32'`, `'float64'`.

    Returns:
        str: the current default float type.
    """
    return _FLOATX


def set_floatx(value):
    """Sets the default float type.

    Note: It is not recommended to set this to float16 for training,
    as this will likely cause numeric stability issues

    Args:
        value (str): `'float16'`, `'float32'`, or `'float64'`.
    """
    global _FLOATX
    if value not in {"float16", "float32", "float64"}:
        raise ValueError("Unknown floatx type: " + str(value))
    _FLOATX = str(value)


def intx():
    """Returns the default int type, as a string.

    E.g. `'int8'`, `'unit8'`, `'int32'`.

    Returns:
        str: the current default int type.
    """
    return _INTX


def set_intx(value):
    """Sets the default int type.

    Args:
        value (str): default int type to use.
        One of `{'int8', 'uint8', 'int16', 'uint16', 'int32', 'int64'}`
    """
    global _INTX
    if value not in {"int8", "uint8", "int16", "uint16", "int32", "int64"}:
        raise ValueError("Unknown intx type: " + str(value))
    _INTX = str(value)


# accelerator
def get_accelerator() -> AcceleratorType:
    "return type of accelerator used"
    return _ACCELERATOR


def set_accelerator(acc: AcceleratorType):
    "Set accelerator type"
    global _ACCELERATOR
    _ACCELERATOR = acc


# backend
def get_backend() -> BackendType:
    """Return the UniSim backend currently in use."""
    return _BACKEND


def set_backend(backend: BackendType):
    """Set UniSim backend to a given framework.

    Args:
        name: Name of the backend, one of {onnx, tensorflow}. Favors Tensorflow
        if available or on GPU, and defaults to Onnx for CPU inference.

    Notes:
        Backend name check is done is `load_backend.py`,
        see `load_backend.py` for the actual loading code.
    """
    global _BACKEND
    _BACKEND = backend



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from functools import cache
from typing import List, Sequence, Tuple

import numpy as np

PAD = [0.0] * 24


@cache
def char2bin(chr: str) -> List[float]:
    cp = ord(chr)
    v = format(cp, "#026b")
    return [float(b) for b in v[2:]]


@cache
def token2bin(word: str, add_space: bool = True) -> List[List[float]]:
    "convert a word"
    if add_space:
        word += " "
    return [char2bin(c) for c in word]


def binarize_str(
    txt: str,
    docid: int,
    chunk_size: int = 512,
    last_chunk_min_size: int = 256,
):
    # binarize
    bins = [char2bin(c) for c in txt]

    # remove last chunk if too small
    lbin = len(bins)
    last_chunk_size = lbin % chunk_size

    # deal with last chunk
    if lbin > chunk_size and last_chunk_size <= last_chunk_min_size:
        # if too small truncate
        if last_chunk_size:  # avoid doing it if by change perfect cut
            bins = bins[: lbin - last_chunk_size]
    else:
        # else pad
        pad_len = chunk_size - len(bins) % chunk_size
        padding = [PAD for _ in range(pad_len)]
        bins.extend(padding)

    # arrayifying and reshaping
    arr = np.array(bins, dtype="float32")
    num_chunks = arr.shape[0] // chunk_size
    arr = np.reshape(arr, (num_chunks, chunk_size, 24))
    return arr, num_chunks, docid


def binarizer(
    txts: Sequence[str],
    chunk_size: int = 512,
    last_chunk_min_size: int = 256,
) -> Tuple[np.ndarray, List[List[int]]]:
    """Binarize text using the RETVec chararcter encoder, as detailed in https://arxiv.org/abs/2302.09207.

    Args:
        txts: Input texts to binarize.

        chunk_size: Chunk size in characters to break up the text.

        last_chunk_min_size: Drop last chunk if it has less than `last_chunk_min_size` characters.

    Returns:
        Binarized text which is ready to be embedded using the RETSim embedding model.
    """
    inputs = []
    chunk_ids = []
    chunk_ids_cursor = 0
    for docid, txt in enumerate(txts):
        arr, num_chunks, docid = binarize_str(
            txt=txt, docid=docid, chunk_size=chunk_size, last_chunk_min_size=last_chunk_min_size
        )
        inputs.extend(arr)
        current_chunks_ids = list(range(chunk_ids_cursor, chunk_ids_cursor + num_chunks))
        chunk_ids.append(current_chunks_ids)
        chunk_ids_cursor += num_chunks

    # stack
    binarized_inputs: np.ndarray = np.stack(inputs, axis=0)
    return binarized_inputs, chunk_ids



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from __future__ import annotations

import dataclasses as dc
from typing import Any, List

from .types import Embedding


@dc.dataclass
class Match:
    """Metadata associated with a query match.

    Attributes:
        idx: Idx of the match, as stored in the index.

        data: Input data of the match, if stored.

        rank: Rank of the match with respect to the query similarity, e.g. 0 for most similar match.

        similarity: Cosine similarity from the match to the query.

        embedding: Embedding of the match data.

        is_match: Whether the match and the query are near-duplicate matches.
    """

    idx: int
    data: Any | None = None
    rank: int | None = None
    similarity: float | None = None
    embedding: Embedding | None = None
    is_match: bool = False


@dc.dataclass
class Result:
    """Metadata associated a search result for a particular query.

    Attributes:
        query_idx: Idx of the query in the list of queries passed to UniSim's search function.

        query_data: Input data of the query.

        query_embedding: Embedding of the query data.

        num_matches: Number of near-duplicate matches for query found.

        matches: A list of Match objects corresponding to the near-duplicate matches found for the query.
    """

    query_idx: int
    query_data: Any | None = None
    query_embedding: Embedding | None = None
    num_matches: int = 0
    matches: List[Match] = dc.field(default_factory=lambda: [])


@dc.dataclass
class ResultCollection:
    """Collection of search results returned.

    Attributes:
        total_matches: Total number of near-duplicate matches found across all search queries.

        results: A list of Result objects corresponding to each search query result.
    """

    total_matches: int = 0
    results: List[Result] = dc.field(default_factory=lambda: [])

    def __repr__(self) -> str:
        res = "-[Results Statistics]-\n"
        res += f"Number of Results: {len(self.results)}\n"
        res += f"Total Matches: {self.total_matches}\n"
        return res

    def merge_result_collection(self, other: ResultCollection) -> ResultCollection:
        self.total_matches += other.total_matches
        self.results.extend(other.results)
        return self



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from enum import Enum


class BackendType(Enum):
    """Backend type.

    One of {unknown, onnx, tf}.
    """

    unknown = 0
    onnx = 1
    tf = 2


class AcceleratorType(Enum):
    """Accelerator type

    One of {unknown, cpu, gpu}.
    """

    unknown = 0
    cpu = 1
    gpu = 2


class IndexerType(Enum):
    """Indexer type

    `exact` for exact search index, `approx` for Approximate Nearest Neighbor (ANN) search index.
    """

    exact = 0
    approx = 1


class ModalityType(Enum):
    """Modality type

    One of {multimodal, text, image}.

    NOTE: Only text is supported in the initial v0 release using unisim.TextSim.
    """

    multimodal = 0
    text = 1
    image = 2



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from typing import Sequence

import numpy as np

from ...types import BatchEmbeddings
from ..embedder import Embedder
from .binarizer import binarizer


class TextEmbedder(Embedder):
    """Converts text to embeddings using the RETSim text embedding model.

    The RETSim model embeds texts into 256-dimensional float vectors. The embeddings
    are trained using TensorFlow Similarity to be robust against adversarial attacks.
    The model is able to handle inputs of arbitrary length.
    """

    def __init__(self, batch_size: int, model_id: str = "text/retsim/v1", verbose: int = 0):
        """Initialize a TextEmbedder.

        Args:
            batch_size: Batch size for inference.

            model_id: ID of model, defaults to `text/retsim/v1`.

            verbose: Verbosity mode.
        """
        # model loading is handled in the super
        super().__init__(batch_size=batch_size, model_id=model_id, verbose=verbose)
        self._chunk_size = 512
        self._embedding_size = 256

    @property
    def embedding_size(self) -> int:
        """Returns the embedding size of the text embedding model."""
        return self._embedding_size

    def embed(
        self,
        inputs: Sequence[str],
    ) -> BatchEmbeddings:
        """Compute embeddings for a batch of text inputs.

        Args:
            inputs: Input texts to embed.

        Returns:
            Text embeddings for each text in `inputs`.
        """
        batch, docids = binarizer(inputs, chunk_size=self._chunk_size)  # noqa
        partial_embeddings = self.predict(batch)
        partial_embeddings = np.asanyarray(partial_embeddings)

        # average partial embeddings together into global embeddings
        global_embeddings = []
        for idxs in docids:
            embs = np.take(partial_embeddings, idxs, axis=0)

            # averaging partial embeddings together, if needed
            if len(embs) > 1:
                avg_emb = np.sum(embs, axis=0) / len(embs)
                avg_emb = avg_emb / np.linalg.norm(avg_emb)
            else:
                avg_emb = embs[0]  # (1, 256) shape

            global_embeddings.append(avg_emb)

        return global_embeddings



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

import os
from importlib import reload

import pandas as pd
import pytest

import unisim
from unisim import TextSim
from unisim.config import BackendType, set_backend

backend_type = ["tf", "onnx"]
index_type = ["exact", "approx"]
BATCH_SIZE = 4


def set_test_backend(b):
    # Reload unisim to pick up the new backend
    os.environ["BACKEND"] = b

    if b == "tf":
        set_backend(BackendType.tf)
    elif b == "onnx":
        set_backend(BackendType.onnx)

    reload(unisim)


@pytest.mark.parametrize("backend_type", backend_type, ids=backend_type)
@pytest.mark.parametrize("index_type", index_type, ids=index_type)
def test_textsim_similarity(backend_type, index_type):
    set_test_backend(b=backend_type)
    tsim = TextSim(index_type=index_type, model_id="text/retsim/v1", batch_size=BATCH_SIZE, verbose=True)

    sim = tsim.similarity("I love icecreams", "I love icecreams")
    assert round(sim, 3) == 1

    test_str = "b" * 512 + "a" * 512
    sim = tsim.similarity(test_str, test_str)
    assert round(sim, 3) == 1

    sim = tsim.similarity("doubting dreams", "rough winds do shake the darling buds of may,")
    assert round(sim, 3) == 0.485

    sim = tsim.similarity("this is a test", "This is a test 😀")
    assert round(sim, 3) == 0.967


@pytest.mark.parametrize("backend_type", backend_type, ids=backend_type)
@pytest.mark.parametrize("index_type", index_type, ids=index_type)
def test_textsim_embed(backend_type, index_type):
    set_test_backend(b=backend_type)
    tsim = TextSim(index_type=index_type, model_id="text/retsim/v1", batch_size=BATCH_SIZE)

    embs = tsim.embed(["This is a test"])
    assert embs.shape == (1, 256)

    embs = tsim.embed(["a" * 2000, "test"])
    assert embs.shape == (2, 256)


@pytest.mark.parametrize("backend_type", backend_type, ids=backend_type)
@pytest.mark.parametrize("index_type", index_type, ids=index_type)
def test_textsim_add(backend_type, index_type):
    set_test_backend(b=backend_type)
    tsim = TextSim(index_type=index_type, model_id="text/retsim/v1", batch_size=BATCH_SIZE)

    s1 = "This is a test"
    s2 = "a" * 2000

    tsim.add([s1])
    tsim.add([s1, s2, s2])


@pytest.mark.parametrize("backend_type", backend_type, ids=backend_type)
@pytest.mark.parametrize("index_type", index_type, ids=index_type)
def test_textsim_basic_search_workflow(backend_type, index_type):
    set_test_backend(b=backend_type)
    tsim = TextSim(index_type=index_type, model_id="text/retsim/v1", batch_size=BATCH_SIZE)
    tsim.info()
    tsim.reset_index()

    s1 = "This is a test"
    s2 = "a" * 2000
    s3 = "cookies"
    s4 = "cookies?"
    tsim.add([s1, s2, s3, s4])

    rc = tsim.search([s3], k=1)
    res_0 = rc.results[0]
    tsim.visualize(res_0)

    assert rc.total_matches == 1
    assert res_0.num_matches == 1

    rc = tsim.search([s3, s1], k=5)
    res_0 = rc.results[0]
    res_1 = rc.results[1]
    tsim.visualize(res_0)

    assert rc.total_matches == 3
    assert res_0.num_matches == 2
    assert res_0.query_idx == 0
    assert res_0.query_data == s3
    assert res_0.query_embedding.shape == (256,)

    assert res_1.num_matches == 1
    assert res_0.matches[0].idx == 2
    assert res_0.matches[1].idx == 3
    assert res_1.matches[0].idx == 0
    assert res_1.matches[0].embedding.shape == (256,)


@pytest.mark.parametrize("backend_type", backend_type, ids=backend_type)
@pytest.mark.parametrize("index_type", index_type, ids=index_type)
def test_textsim_match_two_lists(backend_type, index_type):
    set_test_backend(b=backend_type)
    tsim = TextSim(index_type=index_type, model_id="text/retsim/v1", batch_size=BATCH_SIZE)
    queries = ["test", "This is a test", "cookies", "a" * 1024]
    targets = ["this is a test! 😀", "COOKIES", "a" * 1024 + "b" * 256, "test"]

    df = tsim.match(queries=queries, targets=targets)

    expected_df = pd.DataFrame(
        {
            "query": queries,
            "target": [targets[3], targets[0], targets[1], targets[2]],
            "similarity": [1.0000, 0.9371, 0.8950, 1.000],
            "is_match": [True, True, False, True],
        }
    )
    pd.testing.assert_frame_equal(df, expected_df, check_exact=False, atol=1e-4)


@pytest.mark.parametrize("backend_type", backend_type, ids=backend_type)
@pytest.mark.parametrize("index_type", index_type, ids=index_type)
def test_textsim_match_single_list(backend_type, index_type):
    set_test_backend(b=backend_type)
    tsim = TextSim(index_type=index_type, model_id="text/retsim/v1", batch_size=BATCH_SIZE)
    queries = ["test", "This is a test", "this is a test! 😀", "a" * 1024]

    df = tsim.match(queries)

    expected_df = pd.DataFrame(
        {
            "query": queries,
            "target": [queries[1], queries[2], queries[1], queries[0]],
            "similarity": [0.7674, 0.9371, 0.9371, 0.3772],
            "is_match": [False, True, True, False],
        }
    )
    pd.testing.assert_frame_equal(df, expected_df, check_exact=False, atol=1e-4)



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from pathlib import Path
from typing import Any, Sequence, Union

import numpy as np
import tensorflow as tf
from tensorflow import Tensor
from tensorflow.keras import Model

# typing
from ..types import BatchEmbeddings


def cosine_similarity(query_embeddings: BatchEmbeddings, index_embeddings: BatchEmbeddings) -> Tensor:
    """Compute cosine similarity between embeddings using TensorFlow.

    Args:
        query_embeddings: Embeddings of the content to be searched.

        index_embeddings: Embeddings of the indexed content.

    Returns:
        Matrix of cosine similarity values.
    """
    return tf.matmul(query_embeddings, index_embeddings, transpose_b=True)


def load_model(path: Path, verbose: int = 0):
    """Helper function to load TensorFlow/Keras model.

    Args:
        path: Path to the saved model.

        verbose: Print model summary if verbose.

    Returns:
        tf.keras.Model reloaded from path.
    """
    # specialize path
    mpath = str(path)
    if verbose:
        print(f"|-model path: {mpath}")

    model: tf.keras.Model = tf.keras.models.load_model(mpath)

    if verbose:
        model.summary()
    return model


def predict(model: Model, batch: Union[Sequence[Any], np.ndarray]) -> BatchEmbeddings:
    """Predict using TensorFlow model.

    Args:
        model: TensorFlow/Keras model.

        batch: Batch of inputs to generate predictions.

    Returns:
        Batch of embeddings generated from inputs.
    """
    return model(batch, training=False)



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

import os

from ..config import get_accelerator, get_backend, set_accelerator, set_backend
from ..enums import AcceleratorType, BackendType

TF_AVAILABLE = False

# override existing setting to allow reload
if "BACKEND" in os.environ:
    bs = os.environ["BACKEND"]
    if bs == "onnx":
        set_backend(BackendType.onnx)
    elif bs == "tf":
        set_backend(BackendType.tf)
        import tensorflow as tf
    else:
        raise ValueError(f"Unknown environment backend {bs}")
elif not get_backend() or get_backend() == BackendType.unknown:
    # check if we find tensorflow
    try:
        import tensorflow as tf  # noqa: F403, F401

        TF_AVAILABLE = True
    except ImportError:
        TF_AVAILABLE = False

# choose backend if not set by user
accel = get_accelerator()
backend = get_backend()

if "BACKEND" not in os.environ:
    if backend != BackendType.unknown:
        # backend forced by the user
        pass
    elif accel == AcceleratorType.cpu:
        # on CPU always onnx
        set_backend(BackendType.onnx)
    elif TF_AVAILABLE:
        # on GPU use TF by default
        set_backend(BackendType.tf)
    else:
        # if TensorFlow not available
        set_backend(BackendType.onnx)

# post detection
if get_backend() == BackendType.onnx:
    import onnxruntime as rt

    from .onnx import *  # noqa: F403, F401

    # FIXME onnx accelerator type support
    if rt.get_device() == "GPU":
        set_accelerator(AcceleratorType.gpu)
    else:
        set_accelerator(AcceleratorType.cpu)

elif get_backend() == BackendType.tf:
    from .tf import *  # type: ignore # noqa: F403, F401

    # ensure we are not running out of memory
    gpus = tf.config.list_physical_devices("GPU")
    if gpus:
        for gpu in gpus:
            try:
                tf.config.experimental.set_memory_growth(gpu, True)
            except RuntimeError as e:
                # Memory growth must be set before GPUs have been initialized
                print(e)

else:
    raise ValueError(f"Unknown backend {get_backend()}")

print("INFO: Loaded backend")
print(f"INFO: Using {get_backend().name.upper()} with {get_accelerator().name.upper()}")



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from __future__ import annotations

from typing import Union

from jaxtyping import Float32
from numpy import ndarray
from tensorflow import Tensor

Array = Union[Tensor, ndarray]

# Embeddings
Embedding = Float32[Array, "embedding"]
BatchEmbeddings = Float32[Array, "batch embeddings"]



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from .text_embedder import TextEmbedder  # noqa: F401



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from pathlib import Path
from typing import Any, Dict, Sequence, Union

import numpy as np
from onnxruntime import InferenceSession

from ..types import BatchEmbeddings


def cosine_similarity(query_embeddings: BatchEmbeddings, index_embeddings: BatchEmbeddings) -> np.ndarray:
    """Compute cosine similarity between embeddings using numpy.

    Args:
        query_embeddings: Embeddings of the content to be searched.

        index_embeddings: Embeddings of the indexed content.

    Returns:
        Matrix of cosine similarity values.
    """
    similarity: np.ndarray = np.dot(query_embeddings, index_embeddings.T)
    return similarity


# ONNX providers
_providers = ["CPUExecutionProvider"]


def load_model(path: Path, verbose: int = 0) -> Dict[str, Any]:
    """Helper function to load Onnx model.

    Args:
        path: Path to the saved .onnx model.

        verbose: Print model details if verbose.

    Returns:
        Dict containing Onnx model session, input and output.
    """
    # specialize path
    mpath = str(path.with_suffix(".onnx"))
    if verbose:
        print(f"|-model path: {mpath}")

    sess = InferenceSession(mpath, providers=_providers)

    # getting input/output info dynamically
    model_dict = {
        "sess": sess,
        "input_name": sess.get_inputs()[0].name,
        "input_shape": sess.get_inputs()[0].shape,
        "input_type": sess.get_inputs()[0].type,
        "output_name": sess.get_outputs()[0].name,
        "output_shape": sess.get_outputs()[0].shape,
        "output_type": sess.get_outputs()[0].type,
    }

    if verbose:
        print(f'|-input: {model_dict["input_shape"]}, {model_dict["input_type"]}')
        print(f'|-output: {model_dict["output_shape"]}, {model_dict["output_type"]}')

    return model_dict


def predict(model_dict: Dict[str, Any], batch: Union[Sequence[Any], np.ndarray]) -> BatchEmbeddings:
    """Predict using Onnx model dictionary loaded from load_model().

    Args:
        model_dict: Model loaded using load_model().

        batch: Batch of inputs to generate predictions.

    Returns:
        Batch of embeddings generated from inputs
    """
    out = model_dict["sess"].run([model_dict["output_name"]], {model_dict["input_name"]: batch})
    return np.asanyarray(out[0])



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from __future__ import annotations

import pickle
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Sequence

import numpy as np
from usearch.index import BatchMatches, Index, MetricKind, search

from .dataclass import Match, Result, ResultCollection
from .enums import IndexerType
from .types import BatchEmbeddings


class Indexer:
    """Indexing system for UniSim that uses Usearch
    (https://github.com/unum-cloud/usearch) for both exact and approximate
    nearest neighbor (ANN) search. This enables us to efficiently index and
    find the nearest neighbors of embedding vectors.

    We strongly recommend using ANN search if the size of the dataset you
    are indexing or searching over is large, since it will provide a
    substantial speed up. Using exact search guarantees that you return the
    correct nearest neighbors every single time.
    """

    def __init__(
        self,
        embedding_size: int,
        index_type: IndexerType,
        params: Dict[str, Any] = {},
    ):
        """Initialize Indexer for UniSim, which uses the USearch package.

        For more details on USearch, see https://github.com/unum-cloud/usearch.

        Args:
            embedding_size: Size of the embedding vectors to index.

            index_type: Indexer type, either `IndexerType.exact` for exact
                search or `IndexerType.approx` for ANN search.

            params: Additional parameters to be passed into USearch Index,
                only used when index_type is approx, for ANN search. Supported
                dict keys include {metric, dtype, connectivity, expansion_add,
                expansion_search}, please see the USearch Index documentation
                for more detail on what each parameter does.
        """
        self.embedding_size = embedding_size
        self.index_type = index_type
        self.params = params

        # determine if we use ANN or exact search for the index
        self.use_exact = True if index_type == IndexerType.exact else False

        if self.use_exact:
            # exact matching and we need to store the embeddings in memory
            self.embeddings: List = []

        else:
            # initializing the USearch ANN index
            self.index = Index(
                ndim=self.embedding_size,
                metric="ip",
                dtype=params.get("dtype", "f32"),
                connectivity=params.get("connectivity", 16),
                expansion_add=params.get("expansion_add", 128),
                expansion_search=params.get("expansion_search", 64),
            )
        self.idxs: List[int] = []

    def add(self, embeddings: BatchEmbeddings, idxs: List[int]):
        """Add a batch of embeddings to the indexer.

        Args:
            embeddings: Embeddings to add.

            idxs: Indices corresponding to the embeddings to add.
        """
        self.idxs.extend(idxs)

        if self.use_exact:
            self.embeddings.extend(embeddings)
        else:
            keys = np.asanyarray(idxs)
            embs = np.asanyarray(embeddings)
            self.index.add(keys, embs)

    def search(
        self,
        queries: Sequence[Any],
        query_embeddings: BatchEmbeddings,
        similarity_threshold: float,
        k: int,
        drop_closest_match: bool = False,
        return_data: bool = True,
        return_embeddings: bool = True,
        data: Sequence[Any] = [],
    ) -> ResultCollection:
        """Search for and return the k closest matches for a set of queries,
        and mark the ones that are closer than `similarity_threshold` as
        near-duplicate matches.

        Args:
            queries: Input query data.

            query_embeddings: Embeddings corresponding to input queries.

            similarity_threshold: Similarity threshold for near-duplicate match,
                where a query and a search result are considered near-duplicate matches
                if their similarity is higher than `similarity_threshold`.

            k: Number of nearest neighbors to lookup.

            drop_closest_match: If True, remove the closest match before returning
                results. This is used when search queries == indexed set, since
                each query's closest match will be itself if it was already added
                to the index.

            return_data: Whether to return data corresponding to search results.

            return_embeddings: Whether to return embeddings for search results.

            data: Input data to fetch search result data from.

        Returns
            ResultCollection containing the search results.
        """
        if return_data and not data:
            raise ValueError("Can't return data, data is empty")

        # we need to increase k by 1 if we are dropping the closest match
        if drop_closest_match:
            k += 1

        # Using USearch exact search
        if self.use_exact:
            # check how many items to return in case k is larger than index size
            index_size = len(self.embeddings)
            count = k if k < index_size else index_size

            matches_batch = search(
                np.asanyarray(self.embeddings),
                query_embeddings,
                metric=MetricKind.IP,
                count=count,
                exact=self.use_exact,
            )

        # use USearch ANN search
        else:
            index_size = self.index.size
            count = k if k < index_size else index_size
            matches_batch = self.index.search(query_embeddings, count=count)

        # compute matches
        results_map: Dict[int, Result] = {}
        matches_map: Dict[int, Dict[int, Match]] = defaultdict(dict)

        # for single batch lookup Matches is returned instead of BatchMatches
        if not isinstance(matches_batch, BatchMatches):
            matches_batch = [matches_batch]

        for query_idx, matches in enumerate(matches_batch):
            result = Result(query_idx=query_idx)
            if return_data:
                result.query_data = queries[query_idx]

            if return_embeddings:
                result.query_embedding = query_embeddings[query_idx]

            for rank, m in enumerate(matches):
                # drop closest match if set and increase the ranking of matches
                if drop_closest_match:
                    if rank == 0:
                        continue
                    else:
                        rank -= 1

                target_idx = m.key
                similarity = 1 - m.distance
                embedding = None
                if return_embeddings:
                    embedding = self.embeddings[target_idx] if self.use_exact else self.index[target_idx]
                    embedding = np.squeeze(embedding)

                match = Match(idx=target_idx, rank=rank, similarity=similarity, embedding=embedding)

                if return_data:
                    match.data = data[target_idx]

                # is a match?
                if similarity >= similarity_threshold:
                    match.is_match = True
                    result.num_matches += 1

                # save the match in a map to find it back in partial analysis
                matches_map[query_idx][match.idx] = match

            # save in a map to find it back during partial analysis
            results_map[result.query_idx] = result

        # create and return results collection
        results_collection = ResultCollection()
        for result in results_map.values():
            matches = [m for m in matches_map[result.query_idx].values()]
            result.matches = matches
            results_collection.total_matches += result.num_matches
            results_collection.results.append(result)
        return results_collection

    def reset(self):
        """Reset the index."""
        if self.use_exact:
            self.embeddings = []
        else:
            self.index.reset()
        self.idxs = []

    def save(self, path: Path | str):
        """Save the index to disk.

        Args:
            path: directory to save the index
        """
        if self.use_exact:
            with open(self._save_pickle_path(path), "wb") as f:
                pickle.dump((self.embeddings, self.idxs), f)
        else:
            self.index.save(self._save_usearch_path(path))

    def load(self, path: Path | str):
        """Reload index data from save directory and recreate the index
        with the underlying data.

        Args:
            path: Directory where the indexer was saved.

        Returns:
            Initialized indexer from save directory.
        """
        if self.use_exact:
            with open(self._save_pickle_path(path), "rb") as f:
                data = pickle.load(f)
            self.embeddings = data[0]
            self.idxs = data[1]
        else:
            self.index.load(self._save_usearch_path(path))

    def _save_pickle_path(self, path):
        return Path(path) / "index.pickle"

    def _save_usearch_path(self, path):
        return Path(path) / "index.usearch"



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

import os

from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


long_description = read("README.md")

setup(
    name="unisim",
    version=get_version("unisim/__init__.py"),
    description="UniSim: Universal Similarity",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Google",
    author_email="unisim@google.com",
    url="https://github.com/google/unisim",
    license="MIT",
    install_requires=[
        "tabulate",
        "numpy",
        "tqdm",
        "onnx",
        "jaxtyping",
        "onnxruntime-gpu",
        "pandas",
        "tensorflow>=2.11,<2.16",
        "usearch>=2.6.0",
    ],
    extras_require={
        "dev": [
            "datasets",
            "mypy",
            "pytest",
            "flake8",
            "pytest-cov",
            "twine",
            "tabulate",
            "black",
            "isort",
            "tf2onnx",
        ],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    include_package_data=True,
    packages=find_packages(),
)



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

import numpy as np
import pytest

from unisim.indexer import Indexer

index_type = ["exact", "approx"]
EMBEDDING_SIZE = 3


def set_up_test_indexer(index_type):
    indexer = Indexer(
        embedding_size=EMBEDDING_SIZE,
        index_type=index_type,
        params={},
    )
    embs = np.array([[1, 1, 3], [3, 1, 2]], dtype="float32")
    idxs = [0, 1]

    indexer.add(embs, idxs)
    return indexer


@pytest.mark.parametrize("index_type", index_type, ids=index_type)
def test_indexer_add(index_type):
    indexer = set_up_test_indexer(index_type)
    assert indexer.idxs == [0, 1]

    embs = np.array([[1, 1, 3]], dtype="float32")
    idxs = [2]
    indexer.add(embs, idxs)
    assert indexer.idxs == [0, 1, 2]


@pytest.mark.parametrize("index_type", index_type, ids=index_type)
def test_indexer_search_match(index_type):
    indexer = set_up_test_indexer(index_type)

    query_embs = np.array([[1, 1, 1], [0, 0, 0]], dtype="float32")

    result_collection = indexer.search(
        queries=["query1", "query2"],
        query_embeddings=query_embs,
        similarity_threshold=0.5,
        k=2,
        return_data=True,
        data=["test1", "test2"],
    )
    assert result_collection.total_matches == 2

    result = result_collection.results[0]
    assert result.query_idx == 0
    assert result.query_data == "query1"
    assert result.num_matches == 2
    assert len(result.matches) == 2

    match = result.matches[0]
    assert match.idx == 1
    assert match.rank == 0
    assert match.is_match
    assert match.data == "test2"


@pytest.mark.parametrize("index_type", index_type, ids=index_type)
def test_indexer_search_non_match(index_type):
    indexer = set_up_test_indexer(index_type)

    query_embs = np.array([[0, 0, 0]], dtype="float32")

    result_collection = indexer.search(
        queries=["query1"],
        query_embeddings=query_embs,
        similarity_threshold=0.5,
        k=1,
        return_data=True,
        data=["test1", "test2"],
    )
    assert result_collection.total_matches == 0

    result = result_collection.results[0]
    assert result.query_idx == 0
    assert result.query_data == "query1"
    assert result.num_matches == 0


@pytest.mark.parametrize("index_type", index_type, ids=index_type)
def test_indexer_reset(index_type):
    indexer = set_up_test_indexer(index_type)
    indexer.reset()
    assert indexer.idxs == []

    if indexer.use_exact:
        assert indexer.embeddings == []
    else:
        assert indexer.index.size == 0


@pytest.mark.parametrize("index_type", index_type, ids=index_type)
def test_indexer_save_reload(tmpdir, index_type):
    indexer = set_up_test_indexer(index_type)
    indexer.save(tmpdir)
    indexer.load(tmpdir)

    embs = np.array([[1, 1, 3]], dtype="float32")
    idxs = [2]
    indexer.add(embs, idxs)
    assert indexer.idxs == [0, 1, 2]

    if indexer.use_exact:
        assert len(indexer.embeddings) == 3
    else:
        assert indexer.index.size == 3



# Copyright 2023 Google LLC
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

# -- shared function --
# ! don't import loader-related functions here - will break
# ! instead, if needed (like set_backend()) put them in root __init__.py
from ..config import floatx  # noqa: F401
from ..config import intx  # noqa: F401
from ..config import set_floatx  # noqa: F401
from ..config import set_intx  # noqa: F401

# -- backend specific --
from .load_backend import cosine_similarity, load_model, predict  # noqa: F401
